import pytest
from textwrap import dedent
from treetime.vcf_utils import read_vcf, write_vcf
from treetime import TreeTimeError

def create_vcf(dir, content):
    d = dir / 'data'
    d.mkdir()
    fname = d / "no-header.vcf"
    with open(fname, 'w') as fh:
        print(dedent(content), file=fh)
    return str(fname)

class TestMalformedVcf:

    def test_no_header(self, tmp_path):
        with pytest.raises(TreeTimeError):
            read_vcf(create_vcf(tmp_path, """\
                ##fileformat=VCFv4.3
                1\t5\t.\tT\tC\t.\t.\t.\tGT\t1\t1\t1
            """))

    def test_meta_info_after_header(self, tmp_path):
        with pytest.raises(TreeTimeError):
            read_vcf(create_vcf(tmp_path, """\
                #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tsample_A\tsample_B\tsample_C
                ##fileformat=VCFv4.3
                1\t5\t.\tT\tC\t.\t.\t.\tGT\t1\t1\t1
            """))
    
    def test_inconsistent_sample_numbers(self, tmp_path):
        with pytest.raises(TreeTimeError):
            read_vcf(create_vcf(tmp_path, """\
                ##fileformat=VCFv4.3
                #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tsample_A\tsample_B\tsample_C
                1\t5\t.\tT\tC\t.\t.\t.\tGT\t1\t1\t1
                1\t6\t.\tT\tC\t.\t.\t.\tGT\t1\t1
            """))

    def test_no_data_lines(self, tmp_path):
        with pytest.raises(TreeTimeError):
            read_vcf(create_vcf(tmp_path, """\
                ##fileformat=VCFv4.3
                #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tsample_A\tsample_B\tsample_C
            """))

    def test_consistent_ploidy(self, tmp_path):
        """
        The spec heavily implies that ALT alleles must be supplied for each copy
        of the chromosome, even for no-calls: "If a call cannot be made for a
        sample at a given locus, '.' must be specified for each missing allele
        in the GT field (for example './.' for a diploid genotype and '.' for
        haploid genotype)."
        """
        with pytest.raises(TreeTimeError):
            read_vcf(create_vcf(tmp_path, """\
                ##fileformat=VCFv4.3
                #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tsample_A\tsample_B
                1\t5\t.\tT\tC\t.\t.\t.\tGT\t1\t1\t1/1
            """))

class TestMultipleChromosomes:
    """
    These are valid VCFs but TreeTime cannot yet parse them
    """
    def test_multiple_chromosomes(self, tmp_path):
        with pytest.raises(TreeTimeError):
            read_vcf(create_vcf(tmp_path, """\
                ##fileformat=VCFv4.3
                #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tsample_A\tsample_B\tsample_C
                1\t5\t.\tT\tC\t.\t.\t.\tGT\t1\t1\t1
                2\t6\t.\tT\tC\t.\t.\t.\tGT\t1\t1\t1
            """))

def zero_based(idx):
    """Useful as a form of code commentary. Convert idx from 1-based to 0-based"""
    return idx-1

def write_data(tmp_path_factory, sample_names, data_lines, reference_seq, meta_lines=None, reference_name='reference_name'):
    """helper function to create and write a VCF and FASTA file. Returns a tuple of filenames"""
    if meta_lines:
        vcf = "\n".join(meta_lines)+"\n"
    else:
        vcf = dedent("""\
            ##fileformat=VCFv4.3
            ##contig=<ID=1,length=50>
            ##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
        """)
    vcf += "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t" + "\t".join(sample_names) + "\n"
    for line in data_lines:
        vcf += "\t".join(line) + "\n"
    reference=dedent(f"""\
        >{reference_name}
        {reference_seq}
    """)
    dir = tmp_path_factory.mktemp("data")
    vcf_fn = dir / "snps.vcf"
    ref_fn = dir / "reference.fasta"
    with open(vcf_fn, 'w') as fh:
        print(vcf, file=fh)
    with open(ref_fn, 'w') as fh:
        print(reference, file=fh)
    return (str(vcf_fn), str(ref_fn))


class TestSimpleVcf:
    @pytest.fixture(scope="class")
    def data(self, tmp_path_factory):
        """
        Creates the VCF and reference files (in a temporary location) to be used by
        the tests in this class.
        Returns their filenames, as well as information about their contents which
        we can use as comparisons in tests.
        NOTE: this function is only run once - _not_ once per test.
        """
        sample_names = ["sample_A", "sample_B", "sample_C"]
        data_lines = [
            ["1", "5",  ".", "T", "C",   ".", ".", ".", "GT", "1", "1", "1"],
            ["1", "7",  ".", "T", "G",   ".", ".", ".", "GT", "1", "1", "0"],
            ["1", "14", ".", "C", "T",   ".", ".", ".", "GT", "1", "1", "0"],
            ["1", "18", ".", "C", "T",   ".", ".", ".", "GT", "0", "0", "1"],
            ["1", "28", ".", "A", "N",   ".", ".", ".", "GT", "0", "0", "1"],
            ["1", "29", ".", "A", "N",   ".", ".", ".", "GT", "0", "0", "1"],
            ["1", "30", ".", "A", "N",   ".", ".", ".", "GT", "0", "0", "1"],
            ["1", "33", ".", "A", "C,G", ".", ".", ".", "GT", "1", "2", "0"],
            ["1", "39", ".", "C", "T",   ".", ".", ".", "GT", "1", "0", "0"],
            ["1", "42", ".", "G", "A",   ".", ".", ".", "GT", "0", "1", "0"],
            ["1", "43", ".", "A", "T",   ".", ".", ".", "GT", "1", "1", "0"],
        ]
        positions = [int(line[1]) for line in data_lines]
        reference_seq = 'AAAAAAAAAATGCCCTGCGGGTAAAAAAAAAAAAACTACTTGACCATAAA'
        filenames = write_data(tmp_path_factory, sample_names, data_lines, reference_seq)
        return {
            "filenames": filenames,
            "positions": positions,
            "data_lines": data_lines,
            "sample_names": sample_names,
            "reference_seq": reference_seq,
        }

    def test_mutation_structure(self, data):
        vcf_data = read_vcf(*data['filenames'])
        # The structure of insertions & sequences is a dict a key for each sample, irregardless of whether
        # any insertions/mutations are defined for that sample.
        # samples = {'sample_A', 'sample_B', 'sample_C'}
        assert(set(data['sample_names'])==set(vcf_data['sequences'].keys()))
        assert(set(data['sample_names'])==set(vcf_data['insertions'].keys()))

    def test_mutation_parsing(self, data):
        """
        Test that the correct SNPs (ALTs) are extracted from the VCF for the three samples defined in the VCF
        """
        vcf_data = read_vcf(*data['filenames'])
        def summarise(sample):
            # uses 1-based position so we can easily compare with the VCF file
            return ",".join([f"{pos+1}{alt}" for pos,alt in vcf_data['sequences'][sample].items()])

        assert("5C,7G,14T,33C,39T,43T"==summarise('sample_A'))
        assert("5C,7G,14T,33G,42A,43T"==summarise('sample_B'))
        assert("5C,18T,28N,29N,30N"==summarise('sample_C'))

    def test_no_insertions_parsed(self, data):
        vcf_data = read_vcf(*data['filenames'])
        for sample, ins in vcf_data['insertions'].items():
            assert(ins=={})

    def test_reference_sequence(self, data):
        vcf_data = read_vcf(*data['filenames'])
        assert(data['reference_seq']==vcf_data['reference'])

    def test_positions_with_mutations(self, data):
        vcf_data = read_vcf(*data['filenames'])
        assert(data['positions'] == [pos+1 for pos in vcf_data['positions']]) 


class TestNoCallsOrMissing:
    """
    Tests a few overlapping concepts:
    - The genotype field (i.e. within the column of a sample name) may be "." which results in a
      no call allele. So we call this as SNP(s) changing the REF to N(s). The 4.2 spec refers
      to this as "a call cannot be made for a sample at a given locus". Spec 4.3 defines this
      as "If a call cannot be made for a sample at a given locus, '.' must be specified for each
      missing allele in the GT field (for example ./. for a diploid genotype and . for haploid
      genotype)"
    - The ALT allele may be "*". v4.2. defines this as "missing due to a upstream deletion (sic)"
      and 4.3 defines this as "allele missing due to overlapping deletion". Either way we encode
      this similarly to the no-call above, i.e. changing the REF to N(s). Note that the allele
      must be the one-character "*" - this can't appear within other bases.
    - The ALT allele may be ".". This was introduced in version 4.3 as "a MISSING value (no variant)"
      So we treat this the same as "*" above. Again, this must be the entire allele, '.' can't appear
      as part of other bases. Note that while this doesn't exist in v4.2, it is commonly found in
      v4.2 VCF files, e.g. those produced by `snp_sites`.
    """

    @pytest.fixture(scope="class")
    def data(self, tmp_path_factory):
        reference="ATCGA"
        sample_names = ["sample_A", "sample_B", "sample_C", "sample_D"]
        data_lines = [
            ## No call alleles
            ["1", "2",  ".", "T", "A",    ".", ".", ".", "GT", ".", "1", "0", "0"], # sample A has T2N
            ["1", "4",  ".", "GA", "C",   ".", ".", ".", "GT", ".", "1", "0", "0"], # sample A has GA -> NN
            ## star alleles & dot alleles indicate missing
            ["1", "3",  ".", "C", "*,G",  ".", ".", ".", "GT", "1", "2", "0", "0"], # sample A has C->N
            ["1", "3",  ".", "C", "T,.",  ".", ".", ".", "GT", "0", "0", "1", "2"], # sample D has C->N
            ["1", "4",  ".", "GA", ".,*", ".", ".", ".", "GT", "0", "0", "1", "2"], # both samples C & D have G->N and A->N
        ]
        filenames = write_data(tmp_path_factory, sample_names, data_lines, reference)
        return {"filenames": filenames}

    def test_no_call_allele(self, data):
        vcf_data = read_vcf(*data['filenames'])
        assert(vcf_data['sequences']['sample_A'][zero_based(2)]=='N')
        assert(vcf_data['sequences']['sample_A'][zero_based(4)]=='N')
        assert(vcf_data['sequences']['sample_A'][zero_based(5)]=='N')
        assert(vcf_data['sequences']['sample_B'][zero_based(4)]=='C') 
        assert(vcf_data['sequences']['sample_B'][zero_based(5)]=='-')

    def test_star_allele(self, data):
        vcf_data = read_vcf(*data['filenames'])
        assert(vcf_data['sequences']['sample_A'][zero_based(3)]=='N')
        assert(vcf_data['sequences']['sample_B'][zero_based(3)]=='G')

        assert(vcf_data['sequences']['sample_D'][zero_based(4)]=='N')
        assert(vcf_data['sequences']['sample_D'][zero_based(5)]=='N')

    def test_dot_allele(self, data):
        vcf_data = read_vcf(*data['filenames'])
        assert(vcf_data['sequences']['sample_C'][zero_based(3)]=='T')
        assert(vcf_data['sequences']['sample_D'][zero_based(3)]=='N')

        assert(vcf_data['sequences']['sample_C'][zero_based(4)]=='N')
        assert(vcf_data['sequences']['sample_C'][zero_based(5)]=='N')


    def test_malformed_star_allele(self, tmp_path):
        """* must be an entire allele, not together with other bases"""
        with pytest.raises(TreeTimeError):
            read_vcf(create_vcf(tmp_path, """\
                #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tsample_A\tsample_B\tsample_C
                ##fileformat=VCFv4.3
                1\t5\t.\tT\tC*\t.\t.\t.\tGT\t1\t1\t1
            """))

    def test_malformed_dot_allele(self, tmp_path):
        """. must be an entire allele, not together with other bases"""
        with pytest.raises(TreeTimeError):
            read_vcf(create_vcf(tmp_path, """\
                #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tsample_A\tsample_B\tsample_C
                ##fileformat=VCFv4.3
                1\t5\t.\tT\tC.,G\t.\t.\t.\tGT\t1\t1\t1
            """))


class TestVcfSpecExamples:
    """
    Test the examples provided in the VCF specs. Note that the examples
    aren't provided as per-sample VCF files, so there's a bit of interpretation
    required to create the example data. Note that the actual `test_<name>`
    methods are added dynamically after the class definition.
    """
    def create(self, tmp_path, input):
        lines = ["##fileformat=VCFv4.2", # Actual version may differ, but we don't parse this so it doesn't matter (yet)
            "##contig=<ID=1,length=50>",
            "##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">",
            "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t" + "\t".join(input['samples']),
        ]
        for d in input['data']:
            lines.append(f"{d['chrom']}\t{d['pos']}\t.\t{d['ref']}\t{d['alt']}\t.\t.\t.\tGT\t" + "\t".join([str(x) for x in d['values']]))
        return create_vcf(tmp_path, "\n".join(lines))

    @staticmethod
    def add_test(example_data):
        def test(self, tmp_path):
            vcf = read_vcf(self.create(tmp_path, example_data))
            for sample_name, sample_data in example_data['expect_sequences'].items():
                assert(sample_name in vcf['sequences'])
                for snp in sample_data:
                    assert(len(snp)==2)
                    assert(vcf['sequences'][sample_name][zero_based(int(snp[0]))] == snp[1])
                assert(len(vcf['sequences'][sample_name].keys()) == len(sample_data))
            for sample_name, sample_data in example_data['expect_insertions'].items():
                assert(sample_name in vcf['insertions'])
                for ins in sample_data:
                    assert(vcf['insertions'][sample_name][zero_based(int(ins[0]))] == ins[1:])
                assert(len(vcf['insertions'][sample_name].keys()) == len(sample_data))
        return test

"""
-------- comment re: TreeTime's encoding of insertions --------

Give a reference base of 'C' at 1-based position 2, and an
insertion _after_ this of 'A', TreeTime encodes this as an insertion
of 'CA' at (0-based) position 1. This was unexpected to me,
I would have expected an insertion of just 'A'. However I've written
the tests to conform to TreeTime's existing behaviour (version 0.11.1)

---------------------------------------------------------------
"""

vcf_spec_data = {
    'version_4_2': {
        '5_1_1': {
            'samples': ['A', 'B', 'C'],
            'data': [
                {'chrom': 20, 'pos': 3, 'ref': 'C', 'alt': 'G', 'values': [1, 0, 0]},
                {'chrom': 20, 'pos': 2, 'ref': 'TC', 'alt': 'T,TCA', 'values': [0, 1, 2]}
            ],
            # Note that with TC->TCA, there are no sequence changes (T->T, C->C) but one insertion ("A")
            'expect_sequences': {'A': ['3G'], 'B': ['3-']},
            'expect_insertions': {'C': ['3CA']} # see comment above. Actually only a single base "A" insertion
        },
        '5_1_2': {
            'samples': ['A', 'B'],
            'data': [
                {'chrom': 20, 'pos': 2, 'ref': 'TC', 'alt': 'TG,T', 'values': [1,2]}
            ],
            'expect_sequences': {'A': ['3G'], 'B': ['3-']},
            'expect_insertions': {}
        },
        '5_1_3': {
            'samples': ['A', 'B', 'C'],
            ## This example is quite hard to understand for sample A. The alignment provided indicates that
            ## TCG -> T-G is encoded as "ref: TCG, alt: TG". But without aligning the alt to the ref, the only
            ## sane interpretation of this is 2 changes: C->G and G->-. This is what the spec means by
            ## "the molecular equivalence explicitly listed above in the per-base alignment is discarded so the
            ## actual placement of equivalent g isn't retained" (also explained in example 5.2.4)
            ## Similarly, for sample C, the actual event is described as "A base is inserted wrt the reference
            ## sequence" but the way the VCF file reads we are going to have a SNP (G->A) + a insertion (G)
            'data': [
                {'chrom': 20, 'pos': 2, 'ref': 'TCG', 'alt': 'TG,T,TCAG', 'values': [1,2,3]}
            ],
            'expect_sequences': {'A': ['3G', '4-'], 'B': ['3-', '4-'], 'C': ['4A']},
            'expect_insertions': {'C': ['4AG']} # See comment above
        },
        '5_2_1': {
            'samples': ['A'],
            'data': [
                {'chrom': 20, 'pos': 3, 'ref': 'C', 'alt': 'T', 'values': [1]}
            ],
            'expect_sequences': {'A': ['3T']},
            'expect_insertions': {}
        },
        '5_2_2': {
            'samples': ['A'],
            'data': [
                {'chrom': 20, 'pos': 3, 'ref': 'C', 'alt': 'CTAG', 'values': [1]}
            ],
            'expect_sequences': {},
            'expect_insertions': {'A': ['3CTAG']} # See comment above
        },
        '5_2_3': {
            'samples': ['A'],
            'data': [
                {'chrom': 20, 'pos': 2, 'ref': 'TCG', 'alt': 'T', 'values': [1]}
            ],
            'expect_sequences': {'A': ['3-', '4-']},
            'expect_insertions': {}
        },
    }
}
# dynamically create a test for each example in the above data (by adding methods to the class)
for spec_version, spec_data in vcf_spec_data.items():
    for example_key, example_data in spec_data.items():
        setattr(TestVcfSpecExamples, f"test_{spec_version}_example_{example_key}", TestVcfSpecExamples.add_test(example_data))


class TestMutationAndInsertion:
    """
    Tests the situation where a reference base is mutated _and_ there's an insertion
    """

    @pytest.fixture(scope="class")
    def data(self, tmp_path_factory):
        reference="ATCGA"
        sample_names = ["sample_A", "sample_B"]
        data_lines = [
            ["1", "2",  ".", "T", "GA",    ".", ".", ".", "GT", "1", "0"], # sample A has both a C->G mutation _and_ a subsequent "A" insertion
            ["1", "3",  ".", "CGA", "NTTTA,CCGT",   ".", ".", ".", "GT", "1", "2"], # complex! Both samples have multiple mutations + an insertion
        ]
        filenames = write_data(tmp_path_factory, sample_names, data_lines, reference)
        return {"filenames": filenames}

    def test_single_ref_base_mutation_and_insertion(self, data):
        # This case was missed in treetime 0.11.1
        vcf_data = read_vcf(*data['filenames'])
        assert(vcf_data['sequences']['sample_A'][zero_based(2)]=='G')
        assert(vcf_data['insertions']['sample_A'][zero_based(2)]=='GA') # see comment above re: insertion encoding

    def test_multi_ref_base_mutations_and_insertion(self, data):
        # This case was missed in treetime 0.11.1
        vcf_data = read_vcf(*data['filenames'])
        assert(vcf_data['sequences']['sample_A'][zero_based(3)]=='N')
        assert(vcf_data['sequences']['sample_A'][zero_based(4)]=='T')
        assert(vcf_data['sequences']['sample_A'][zero_based(5)]=='T')
        assert(vcf_data['insertions']['sample_A'][zero_based(5)]=='TTA') # see comment above re: insertion encoding

        assert(vcf_data['sequences']['sample_B'][zero_based(4)]=='C')
        assert(vcf_data['sequences']['sample_B'][zero_based(5)]=='G')
        assert(vcf_data['insertions']['sample_B'][zero_based(5)]=='GT') # see comment above re: insertion encoding

class TestMetadataParsing:
        
    def test_simple_haploid(self, tmp_path):
        data = read_vcf(create_vcf(tmp_path, """\
            ##fileformat=VCFv4.3
            #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tsample_A\tsample_B\tsample_C
            1\t5\t.\tT\tC\t.\t.\t.\tGT\t1\t1\t1
            1\t6\t.\tT\tC\t.\t.\t.\tGT\t1\t1\t1
        """))
        assert(data['metadata']['chrom']=='1')
        assert(data['metadata']['ploidy']==1)
        assert(data['metadata']['meta_lines']==['##fileformat=VCFv4.3'])

    def test_simple_diploid(self, tmp_path):
        data = read_vcf(create_vcf(tmp_path, """\
            ##fileformat=VCFv4.3
            ##contig=<ID=1,length=50>
            ##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
            #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tsample_A
            foo\t5\t.\tT\tC\t.\t.\t.\tGT\t1/1
            foo\t6\t.\tT\tC\t.\t.\t.\tGT\t.|.
        """))
        assert(data['metadata']['chrom']=='foo')
        assert(data['metadata']['ploidy']==2)
        assert(data['metadata']['meta_lines']==['##fileformat=VCFv4.3', '##contig=<ID=1,length=50>', '##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">'])



    
def roundtrip(tmp_path, sample_names, data_lines, reference, meta_lines, pass_metadata=True):
    """
    Write the provided data to a (temporary) VCF file.
    Then read it via `read_vcf` and store this as `vcf_a`.
    Then write this data via `write_vcf` and read it back, storing as `vcf_b`.
    Returns a tuple of (vcf_a, vcf_b)
    """

    dir = tmp_path / 'data'
    dir.mkdir()
    vcf_filename_a = str(dir / "a.vcf")
    vcf_filename_b = str(dir / "b.vcf")
    reference_filename = str(dir / 'reference.fasta')

    with open(str(dir / 'reference.fasta'), 'w') as fh:
        print(dedent(f"""\
            >reference_name
            {reference}
        """), file=fh)

    with open(vcf_filename_a, 'w') as fh:
        vcf_lines = meta_lines[:] + \
            ["#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t" + "\t".join(sample_names)] + \
            ["\t".join(data) for data in data_lines]
        print("\n".join(vcf_lines), file=fh)

    vcf_a = read_vcf(vcf_filename_a, reference_filename)

    ## take this data and write it out as a VCF
    tree_dict = {'sequences': vcf_a['sequences'], 'reference': vcf_a['reference'], 'positions': vcf_a['positions']}
    if pass_metadata:
        tree_dict['metadata'] = vcf_a['metadata']
    write_vcf(tree_dict, vcf_filename_b)

    ## then read in this (just-created) VCF
    vcf_b = read_vcf(vcf_filename_b, reference_filename)

    ## uncomment the following & run pytest with "-s" to see the contents of the VCF file being written by TreeTime
    # print("_________________________________")
    # with open(vcf_filename_b) as fh:
    #     for line in fh:
    #         print(line, end='')
    # print("\n_________________________________")

    return (vcf_a, vcf_b)


class TestWriting:
    """
    Write a simple VCF file out (created by hand), parse it with `read_vcf` (called "vcf_a")
    then write that data out and read it back in (called "vcf_b"). vcf_a should equal vcf_b
    for the data we care about.
    """

    def test_basic_roundtripping_vcf(self, tmp_path):
        [vcf_a, vcf_b] = roundtrip(tmp_path,
            reference = "ATCGACC",
            meta_lines = [
                "##fileformat=VCFv4.3",
                "##contig=<ID=1,length=7>",
                "##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">"
            ],
            sample_names = ["sample_A", "sample_B"],
            data_lines = [
                ["foo", "2",  ".", "T", "G",    ".", ".", ".", "GT", "1", "0"],
                ["foo", "3",  ".", "C", "GA,*", ".", ".", ".", "GT", "1", "2"], # "A" insertion will be lost on roundtrip!
                ["foo", "5",  ".", "A", "TC",   ".", ".", ".", "GT", ".", "1"],
                ["foo", "6",  ".", "CC", "C",   ".", ".", ".", "GT", "1", "0"], # del of (1-based) pos 7 in sample A
            ],
            pass_metadata=False
        )

        ## reference is certainly the same, the same FASTA file is being used for both read_vcf commands, but check anyway
        assert(vcf_a['reference']==vcf_a['reference'])
        ## insertions, if there are any, _won't_ be the same because `write_vcf` doesn't read insertions even if they're provided
        ## as input. Note that vcf_a does have an insertion!

        ## The meta-lines are different (as expected) and since we are not passing the metadata information to `write_vcf`
        ## we get back the defaults, which differ from the input
        assert(vcf_a['metadata']['chrom'] == "foo" and vcf_b['metadata']['chrom'] == "1")
        assert(vcf_a['metadata']['ploidy'] == 1 and vcf_b['metadata']['ploidy'] == 2)

        ## positions should be the same (positions are only reflective of data in `sequences`, so the ignoring of insertions is ok)
        assert(vcf_a['positions']==vcf_a['positions'])
        ## check sequences are the same
        assert(vcf_a['sequences']==vcf_b['sequences'])

    def test_basic_roundtripping_vcf_with_metadata(self, tmp_path):
        [vcf_a, vcf_b] = roundtrip(tmp_path,
            reference = "ATCGACC",
            meta_lines = [
                "##fileformat=VCFv4.3",
                "##contig=<ID=1,length=7>",
                "##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">"
            ],
            sample_names = ["sample_A", "sample_B"],
            data_lines = [
                ["1", "2",  ".", "T", "G",    ".", ".", ".", "GT", "1", "0"],
                ["1", "3",  ".", "C", "GA,*", ".", ".", ".", "GT", "1", "2"], # "A" insertion will be lost on roundtrip!
                ["1", "5",  ".", "A", "TC",   ".", ".", ".", "GT", ".", "1"],
                ["1", "6",  ".", "CC", "C",   ".", ".", ".", "GT", "1", "0"], # del of (1-based) pos 7 in sample A
            ],
            pass_metadata=True
        )

        ## reference is certainly the same, the same FASTA file is being used for both read_vcf commands, but check anyway
        assert(vcf_a['reference']==vcf_a['reference'])
        ## insertions, if there are any, _won't_ be the same because `write_vcf` doesn't read insertions even if they're provided
        ## as input. Note that vcf_a does have an insertion!

        ## The meta-lines are different (as expected) but the chrom + ploidy should be the same
        assert(vcf_a['metadata']['chrom'] == vcf_b['metadata']['chrom'])
        assert(vcf_a['metadata']['ploidy'] == vcf_b['metadata']['ploidy'])

        ## positions should be the same (positions are only reflective of data in `sequences`, so the ignoring of insertions is ok)
        assert(vcf_a['positions']==vcf_a['positions'])
        ## check sequences are the same
        assert(vcf_a['sequences']==vcf_b['sequences'])



import os, sys
from Bio import AlignIO, Phylo
from .vcf_utils import read_vcf, write_vcf
from .seq_utils import alphabets
from Bio import __version__ as bioversion
from . import version as treetime_version
import numpy as np

def get_outdir(params, suffix='_treetime'):
    if params.outdir:
        if os.path.exists(params.outdir):
            if os.path.isdir(params.outdir):
                return params.outdir.rstrip('/') + '/'
            else:
                print("designated output location %s is not a directory"%params.outdir, file=sys.stderr)
        else:
            os.makedirs(params.outdir)
            return params.outdir.rstrip('/') + '/'

    from datetime import datetime
    outdir_stem = datetime.now().date().isoformat()
    outdir = outdir_stem + suffix.rstrip('/')+'/'
    count = 1
    while os.path.exists(outdir):
        outdir = outdir_stem + '-%04d'%count + suffix.rstrip('/')+'/'
        count += 1

    os.makedirs(outdir)
    return outdir

def get_basename(params, outdir):
    # if params.aln:
    #     basename = outdir + '.'.join(params.aln.split('/')[-1].split('.')[:-1])
    # elif params.tree:
    #     basename = outdir + '.'.join(params.tree.split('/')[-1].split('.')[:-1])
    # else:
    basename = outdir
    return basename

def read_in_DRMs(drm_file, offset):
    import pandas as pd

    DRMs = {}
    drmPositions = []

    df = pd.read_csv(drm_file, sep='\t')
    for mi, m in df.iterrows():
        pos = m.GENOMIC_POSITION-1+offset #put in correct numbering
        drmPositions.append(pos)

        if pos in DRMs:
            DRMs[pos]['alt_base'][m.ALT_BASE] = m.SUBSTITUTION
        else:
            DRMs[pos] = {}
            DRMs[pos]['drug'] = m.DRUG
            DRMs[pos]['alt_base'] = {}
            DRMs[pos]['alt_base'][m.ALT_BASE] = m.SUBSTITUTION
            DRMs[pos]['gene'] = m.GENE

    drmPositions = np.array(drmPositions)
    drmPositions = np.unique(drmPositions)
    drmPositions = np.sort(drmPositions)

    DRM_info = {'DRMs': DRMs,
            'drmPositions': drmPositions}

    return DRM_info


def read_if_vcf(params):
    """
    Checks if input is VCF and reads in appropriately if it is
    """
    ref = None
    aln = params.aln
    fixed_pi = None
    if hasattr(params, 'aln') and params.aln is not None:
        if any([params.aln.lower().endswith(x) for x in ['.vcf', '.vcf.gz']]):
            if not params.vcf_reference:
                print("ERROR: a reference Fasta is required with VCF-format alignments")
                return -1
            compress_seq = read_vcf(params.aln, params.vcf_reference)
            sequences = compress_seq['sequences']
            ref = compress_seq['reference']
            aln = sequences

            if not hasattr(params, 'gtr') or params.gtr=="infer": #if not specified, set it:
                alpha = alphabets['aa'] if params.aa else alphabets['nuc']
                fixed_pi = [ref.count(base)/len(ref) for base in alpha]
                if fixed_pi[-1] == 0:
                    fixed_pi[-1] = 0.05
                    fixed_pi = [v-0.01 for v in fixed_pi]

    return aln, ref, fixed_pi


def plot_rtt(tt, fname):
    tt.plot_root_to_tip()

    from matplotlib import pyplot as plt
    plt.savefig(fname)
    print("--- root-to-tip plot saved to  \n\t"+fname)


def export_sequences_and_tree(tt, basename, is_vcf=False, zero_based=False,
                              report_ambiguous=False, timetree=False, confidence=False,
                              reconstruct_tip_states=False, tree_suffix=''):
    seq_info = is_vcf or tt.aln
    if is_vcf:
        outaln_name = basename + f'ancestral_sequences{tree_suffix}.vcf'
        write_vcf(tt.get_reconstructed_alignment(reconstruct_tip_states=reconstruct_tip_states), outaln_name)
    elif tt.aln:
        outaln_name = basename + f'ancestral_sequences{tree_suffix}.fasta'
        AlignIO.write(tt.get_reconstructed_alignment(reconstruct_tip_states=reconstruct_tip_states), outaln_name, 'fasta')
    if seq_info:
        print("\n--- alignment including ancestral nodes saved as  \n\t %s\n"%outaln_name)

    # decorate tree with inferred mutations
    terminal_count = 0
    offset = 0 if zero_based else 1
    if timetree:
        dates_fname = basename + f'dates{tree_suffix}.tsv'
        fh_dates = open(dates_fname, 'w', encoding='utf-8')
        if confidence:
            fh_dates.write('#Lower and upper bound delineate the 90% max posterior region\n')
            fh_dates.write('#node\tdate\tnumeric date\tlower bound\tupper bound\n')
        else:
            fh_dates.write('#node\tdate\tnumeric date\n')

    mutations_out = open(basename + "branch_mutations.txt", "w")
    mutations_out.write("node\tstate1\tpos\tstate2\n")
    for n in tt.tree.find_clades():
        if timetree:
            if confidence:
                if n.bad_branch:
                    fh_dates.write('%s\t--\t--\t--\t--\n'%(n.name))
                else:
                    conf = tt.get_max_posterior_region(n, fraction=0.9)
                    fh_dates.write('%s\t%s\t%f\t%f\t%f\n'%(n.name, n.date, n.numdate,conf[0], conf[1]))
            else:
                if n.bad_branch:
                    fh_dates.write('%s\t--\t--\n'%(n.name))
                else:
                    fh_dates.write('%s\t%s\t%f\n'%(n.name, n.date, n.numdate))

        n.confidence=None
        # due to a bug in older versions of biopython that truncated filenames in nexus export
        # we truncate them by hand and make them unique.
        if n.is_terminal() and len(n.name)>40 and bioversion<"1.69":
            n.name = n.name[:35]+'_%03d'%terminal_count
            terminal_count+=1
        n.comment=''
        if seq_info and len(n.mutations):
            if n.mask is None:
                if report_ambiguous:
                    n.comment= '&mutations="' + ','.join([a+str(pos + offset)+d for (a,pos, d) in n.mutations])+'"'
                else:
                    n.comment= '&mutations="' + ','.join([a+str(pos + offset)+d for (a,pos, d) in n.mutations
                                                        if tt.gtr.ambiguous not in [a,d]])+'"'
            else:
                if report_ambiguous:
                    n.comment= '&mutations="' + ','.join([a+str(pos + offset)+d for (a,pos, d) in n.mutations if n.mask[pos]>0])+f'",mcc="{n.mcc}"'
                else:
                    n.comment= '&mutations="' + ','.join([a+str(pos + offset)+d for (a,pos, d) in n.mutations
                                                        if tt.gtr.ambiguous not in [a,d] and n.mask[pos]>0])+f'",mcc="{n.mcc}"'

            for (a, pos, d) in n.mutations:
                if tt.gtr.ambiguous not in [a,d] or report_ambiguous:
                    mutations_out.write("%s\t%s\t%s\t%s\n" %(n.name, a, pos + 1, d))
        if timetree:
            n.comment+=(',' if n.comment else '&') + 'date=%1.2f'%n.numdate
    mutations_out.close()

    # write tree to file
    fmt_bl = "%1.7f" if tt.data.full_length<1e6 else "%1.9e"
    if timetree:
        outtree_name = basename + f'timetree{tree_suffix}.nexus'
        print("--- saved divergence times in \n\t %s\n"%dates_fname)
        Phylo.write(tt.tree, outtree_name, 'nexus')
    else:
        outtree_name = basename + f'annotated_tree{tree_suffix}.nexus'
        Phylo.write(tt.tree, outtree_name, 'nexus', format_branch_length=fmt_bl)
    print("--- tree saved in nexus format as  \n\t %s\n"%outtree_name)

    # Only create auspice json if there is sequence information
    auspice = create_auspice_json(tt, timetree=timetree, confidence=confidence, seq_info=seq_info)
    outtree_name_json = basename + f'auspice_tree{tree_suffix}.json'
    with open(outtree_name_json, 'w') as fh:
        import json
        json.dump(auspice, fh, indent=0)
        print("--- tree saved in auspice json format as  \n\t %s\n"%outtree_name_json)

    if timetree:
        for n in tt.tree.find_clades():
            n.branch_length = n.mutation_length
        outtree_name = basename + f'divergence_tree{tree_suffix}.nexus'
        Phylo.write(tt.tree, outtree_name, 'nexus', format_branch_length=fmt_bl)
        print("--- divergence tree saved in nexus format as  \n\t %s\n"%outtree_name)

    if hasattr(tt, 'outliers') and tt.outliers is not None:
        print("--- saved detected outliers as " + basename + 'outliers.tsv')
        tt.outliers.to_csv(basename + 'outliers.tsv', sep='\t')

def print_save_plot_skyline(tt, n_std=2.0, screen=True, save='', plot='', gen=50):
    if plot:
        import matplotlib.pyplot as plt

    skyline, conf = tt.merger_model.skyline_inferred(gen=gen, confidence=n_std)
    if save: fh = open(save, 'w', encoding='utf-8')
    header1 = "Skyline assuming "+ str(gen)+" gen/year and approximate confidence bounds (+/- %f standard deviations of the LH)\n"%n_std
    header2 = "date \tN_e \tlower \tupper"
    if screen: print('\t'+header1+'\t'+header2)
    if save: fh.write("#"+ header1+'#'+header2+'\n')
    for (x,y, y1, y2) in zip(skyline.x, skyline.y, conf[0], conf[1]):
        if screen: print("\t%1.3f\t%1.3e\t%1.3e\t%1.3e"%(x,y, y1, y2))
        if save: fh.write("%1.3f\t%1.3e\t%1.3e\t%1.3e\n"%(x,y, y1, y2))

    if save:
        print("\n --- written skyline to %s\n"%save)
        fh.close()

    if plot:
        plt.figure()
        plt.fill_between(skyline.x, conf[0], conf[1], color=(0.8, 0.8, 0.8))
        plt.plot(skyline.x, skyline.y, label='maximum likelihood skyline')
        plt.yscale('log')
        plt.legend()
        plt.ticklabel_format(axis='x',useOffset=False)
        plt.savefig(plot)



def create_auspice_json(tt, timetree=False, confidence=False, seq_info=False):
    # mock up meta data for auspice json
    from datetime import datetime
    meta = {
        "title": f"Auspice visualization of TreeTime (v{treetime_version}) analysis",
        "build_url": "https://github.com/neherlab/treetime",
        "last_updated": datetime.now().strftime("%Y-%m-%d"),
        "treetime_version": treetime_version,
        "genome_annotations": {
            "nuc":{"start":1, "end":int(tt.data.full_length), "type":"source", "strand":"+:"}
        },
        "panels":["tree", "entropy"],
        "colorings": [
            {
                "title": "Date",
                "type": "continuous",
                "key": "num_date",
            },
            {
                "title": "Genotype",
                "type": "categorical",
                "key": "gt",
            },
            {
                "title": "Excluded",
                "type": "categorical",
                "key": "bad_branch"
            },
            {
                "title": "Branch Support",
                "type": "continuous",
                "key": "confidence"
            }
        ],
        "display_defaults": {"color_by":"bad_branch"},
        "filters": ["bad_branch"]
    }

    def node_to_json(n, pdiv=0.0):
        j = {"name":n.name, "node_attrs":{}, "branch_attrs":{}}
        if n.clades:
            j["children"] = []

        if timetree:
            j["node_attrs"]["num_date"] = {"value":float(n.numdate)}
            if confidence:
                conf = tt.get_max_posterior_region(n, fraction=0.9)
                j["node_attrs"]["num_date"]["confidence"] = (float(conf[0]), float(conf[1]))
        j["node_attrs"]["div"] = float(pdiv + n.mutation_length)
        j["node_attrs"]["bad_branch"] = {"value": "Yes" if n.bad_branch else "No"}

        if seq_info: # only add mutations to the json if run with sequence data (fasta or vcf)
            j["branch_attrs"]["mutations"] = {"nuc": [f"{a}{pos+1}{d}" for a,pos,d in n.mutations if d in "ACGT-"]}
            # generate bootstrap confidence substitute via the negative exponential of the number of mutations
            # this is the bootstrap confidence for iid mutations (only ACGT mutations)
            j["node_attrs"]["confidence"] = {"value":round(1-np.exp(-len([pos for a,pos,d in n.mutations if d in "ACGT"])),3)
                                            if not n.is_terminal() else 1.0}
        return j

    # create the tree data structure from the Biopython tree
    tree = node_to_json(tt.tree.root, 0.0)
    # dictionary to look up nodes by name
    node_lookup = {tt.tree.root.name: tree}
    for n in tt.tree.get_nonterminals():
        n_json = node_lookup[n.name]
        for c in n.clades:
            # generate node jsons for all children and attach them the to parent
            n_json["children"].append(node_to_json(c, n_json["node_attrs"]["div"]))
            node_lookup[c.name] = n_json["children"][-1]

    return {"meta":meta, "tree":tree}



import numpy as np
import pandas as pd

def residual_filter(tt, n_iqd):
    terminals = tt.tree.get_terminals()
    clock_rate = tt.clock_model['slope']
    icpt = tt.clock_model['intercept']
    res = {}
    for node in terminals:
        if hasattr(node, 'raw_date_constraint') and  (node.raw_date_constraint is not None):
            res[node] = node.dist2root - clock_rate*np.mean(node.raw_date_constraint) - icpt

    residuals = np.array(list(res.values()))
    iqd = np.percentile(residuals,75) - np.percentile(residuals,25)
    outliers = {}
    for node,r in res.items():
        if abs(r)>n_iqd*iqd and node.up.up is not None:
            node.bad_branch=True
            outliers[node.name] = {'tau':(node.dist2root - icpt)/clock_rate,  'avg_date': np.mean(node.raw_date_constraint),
                                'exact_date': node.raw_date_constraint if type(node) is float else None,
                                'residual': r/iqd}
        else:
            node.bad_branch=False

    tt.outliers=None
    if len(outliers):
        outlier_df = pd.DataFrame(outliers).T.loc[:,['avg_date', 'tau', 'residual']]\
                                .rename(columns={'avg_date':'given_date', 'tau':'apparent_date'})
        tt.logger("Clock_filter.residual_filter marked the following outliers:", 2, warn=True)
        if tt.verbose>=2:
            print(outlier_df)
        tt.outliers = outlier_df
    return len(outliers)

def local_filter(tt, z_score_threshold):
    tt.logger(f"TreeTime.ClockFilter: starting local_outlier_detection", 2)

    node_info = collect_node_info(tt)

    node_info, z_scale = calculate_node_timings(tt, node_info)
    tt.logger(f"TreeTime.ClockFilter: z-scale {z_scale:1.2f}", 2)

    outliers = flag_outliers(tt, node_info, z_score_threshold, z_scale)

    for n in tt.tree.get_terminals():
        if n.name in outliers:
            n.bad_branch = True

    tt.outliers=None
    if len(outliers):
        outlier_df = pd.DataFrame(outliers).T.loc[:,['avg_date', 'tau', 'z', 'diagnosis']]\
                                .rename(columns={'avg_date':'given_date', 'tau':'apparent_date', 'z':'z-score'})
        tt.logger("Clock_filter.local_filter marked the following outliers", 2, warn=True)
        if tt.verbose>=2:
            print(outlier_df)
        tt.outliers = outlier_df
    return len(outliers)


def flag_outliers(tt, node_info, z_score_threshold, z_scale):
    def add_outlier_info(z, n, n_info, parent_tau, mu):
        n_info['z'] = z
        diagnosis=''
        # muts = n_info["nmuts"] if n.is_terminal() else 0.0
        # parent_tau = node_info[n.up.name]['tau'] if n.is_terminal() else n_info['tau']
        if z<0:
            if np.abs(n_info['avg_date']-parent_tau) > n_info["nmuts"]/mu:
                diagnosis='date_too_early'
            else:
                diagnosis = 'excess_mutations'
        else:
            diagnosis = 'date_too_late'
        n_info['diagnosis'] = diagnosis
        return n_info

    outliers = {}
    mu = tt.clock_model['slope']*tt.data.full_length
    for n in tt.tree.get_terminals():
        if n.up.up is None:
            continue # do not label children of the root as bad -- typically a problem with root choice that will be fixed anyway
        n_info = node_info[n.name]
        parent_tau = node_info[n.up.name]['tau']
        if n_info['exact_date']:
            z = (n_info['avg_date'] - n_info['tau'])/z_scale
            if np.abs(z) > z_score_threshold:
                outliers[n.name] = add_outlier_info(z, n, n_info, parent_tau, mu)
        elif n.raw_date_constraint and len(n.raw_date_constraint):
            zs = [(n_info['tau'] - x)/z_scale for x in n.raw_date_constraint]
            if zs[0]*zs[1]>0 and np.min(np.abs(zs))>z_score_threshold:
                outliers[n.name] = add_outlier_info(zs[0] if np.abs(zs[0])<np.abs(zs[1]) else zs[1],
                                                    n, n_info, parent_tau, mu)

    return outliers

def calculate_node_timings(tt, node_info, eps=0.2):
    mu = tt.clock_model['slope']*tt.data.full_length
    sigma_sq = (3/mu)**2
    tt.logger(f"Clockfilter.calculate_node_timings: mu={mu:1.3e}/y, sigma={3/mu:1.3e}y", 2)
    for n in tt.tree.find_clades(order='postorder'):
        p = node_info[n.name]
        if not p['exact_date'] or p['skip']:
            continue

        if n.is_terminal():
            prefactor = (p["observations"]/sigma_sq + mu**2/(p["nmuts"]+eps))
            p["a"] = (p["avg_date"]/sigma_sq + mu*p["nmuts"]/(p["nmuts"]+eps))/prefactor
        else:
            children = [node_info[c.name] for c in n if (not node_info[c.name]['skip']) and node_info[c.name]['exact_date']]
            if n==tt.tree.root:
                tmp_children_1 = mu*np.sum([(mu*c["a"]-c["nmuts"])/(eps+c["nmuts"]) for c in children])
                tmp_children_2 = mu**2*np.sum([(1-c["b"])/(eps+c["nmuts"]) for c in children])
                prefactor = (p["observations"]/sigma_sq + tmp_children_2)
                p["a"] = (p["observations"]*p["avg_date"]/sigma_sq + tmp_children_1)/prefactor
            else:
                tmp_children_1 = mu*np.sum([(mu*c["a"]-c["nmuts"])/(eps+c["nmuts"]) for c in children])
                tmp_children_2 = mu**2*np.sum([(1-c["b"])/(eps+c["nmuts"]) for c in children])
                prefactor = (p["observations"]/sigma_sq + mu**2/(p["nmuts"]+eps) + tmp_children_2)
                p["a"] = (p["observations"]*p["avg_date"]/sigma_sq + mu*p["nmuts"]/(p["nmuts"]+eps)+tmp_children_1)/prefactor
        p["b"] = mu**2/(p["nmuts"]+eps)/prefactor

    node_info[tt.tree.root.name]["tau"] = node_info[tt.tree.root.name]["a"]

    ## need to deal with tips without exact dates below.
    dev = []
    for n in tt.tree.get_nonterminals(order='preorder'):
        p = node_info[n.name]
        for c in n:
            c_info = node_info[c.name]
            if c_info['skip']:
                c_info['tau']=p['tau']
            else:
                if c_info['exact_date']:
                    c_info["tau"] = c_info["a"] + c_info["b"]*p["tau"]
                else:
                    c_info["tau"] = p["tau"] + c_info['nmuts']/mu
            if c.is_terminal() and c_info['exact_date']:
                dev.append(c_info['avg_date']-c_info['tau'])

    return node_info, np.std(dev)


def collect_node_info(tt, percentile_for_exact_date=90):
    node_info = {}
    aln = tt.aln or False
    if aln and (not tt.sequence_reconstruction):
        tt.infer_ancestral_sequences(infer_gtr=False)
    L = tt.data.full_length

    date_uncertainty = [np.abs(n.raw_date_constraint[1]-n.raw_date_constraint[0])
                            if type(n.raw_date_constraint)!=float else 0.0
                        for n in tt.tree.get_terminals()
                            if n.raw_date_constraint is not None]
    from scipy.stats import scoreatpercentile
    uncertainty_cutoff = scoreatpercentile(date_uncertainty, percentile_for_exact_date)*1.01

    for n in tt.tree.get_nonterminals(order='postorder'):
        parent = {"dates": [], "tips": {}, "skip":False}
        exact_dates = 0
        for c in n:
            if c.is_terminal():
                child = {'skip':False}
                child["nmuts"] = len([m for m in c.mutations if m[-1] in 'ACGT']) if aln \
                                      else np.round(c.branch_length*L)
                if c.raw_date_constraint is None:
                    child['exact_date'] = False
                elif type(c.raw_date_constraint)==float:
                    child['exact_date'] = True
                else:
                    child['exact_date'] = np.abs(c.raw_date_constraint[1]-c.raw_date_constraint[0])<=uncertainty_cutoff

                if child['exact_date']:
                    exact_dates += 1
                    if child["nmuts"]==0:
                        child['skip'] = True
                        parent["tips"][c.name]={'date': np.mean(c.raw_date_constraint),
                                                'exact_date':child['exact_date']}
                    else:
                        child['skip'] = False
                        child['observations'] = 1
                if c.raw_date_constraint is not None:
                    child["avg_date"] = np.mean(c.raw_date_constraint)
                node_info[c.name] = child
            else:
                if node_info[c.name]['exact_date']:
                    exact_dates += 1

        parent['exact_date'] = exact_dates>0

        parent["nmuts"] = len([m for m in n.mutations if m[-1] in 'ACGT']) if aln else np.round(n.branch_length*L)
        d = [v['date'] for v in parent['tips'].values() if v['exact_date']]
        parent["observations"] = len(d)
        parent["avg_date"] = np.mean(d) if len(d) else 0.0
        node_info[n.name] = parent

    return node_info



import gzip
import numpy as np
from collections import defaultdict
from textwrap import fill
from . import TreeTimeError

## Functions to read in and print out VCF files

def read_vcf(vcf_file, ref_file=None):
    """
    Reads in a vcf/vcf.gz file and associated
    reference sequence fasta (to which the VCF file is mapped).

    Parses mutations, insertions, and deletions and stores them in a nested dict,
    see 'returns' for the dict structure.

    Calls with heterozygous values 0/1, 0/2, etc and no-calls (./.) are
    replaced with Ns at the associated sites.

    Positions are stored to correspond the location in the reference sequence
    in Python (numbering is transformed to start at 0)

    Parameters
    ----------
    vcf_file : string
        Path to the vcf or vcf.gz file to be read in
    ref_file : string, optional
        Path to the fasta reference file to be read in

    Returns
    --------
    compress_seq : nested dict
        In the format: ::

            {
            'reference':'AGCTCGA..A',
            'sequences': { 'seq1':{4:'A', 7:'-'}, 'seq2':{100:'C'} },
            'insertions': { 'seq1':{4:'ATT'}, 'seq3':{1:'TT', 10:'CAG'} },
            'positions': [1,4,7,10,100...]
            }

        references : string
            String of the reference sequence read from the Fasta, to which
            the variable sites are mapped
        sequences : nested dict
            Dict containing sequence names as keys which map to dicts
            that have position as key and the single-base mutation (or deletion)
            as values
        insertions : nested dict
            Dict in the same format as the above, which stores insertions and their
            locations. The first base of the insertion is the same as whatever is
            currently in that position (Ref if no mutation, mutation in 'sequences'
            otherwise), so the current base can be directly replaced by the bases held here.
        positions : list
            Python list of all positions with a mutation, insertion, or deletion.

    """
    import re
    ALT_CHARS = re.compile(r"^([ACGTNacgtn]+|\*|\.)$") # straight from the VCF 4.3 spec

    #Programming Note:
    # Note on VCF Format
    # -------------------
    # 'Insertion where there are also deletions' (special handling)
    #     Ex:
    #       REF     ALT         Seq1    Seq2
    #       GC      GCC,G       1/1     2/2
    #     Insertions formatted differently - don't know how many bp match
    #     the Ref (unlike simple insert below). Could be mutations, also.
    # 'Deletion'
    #     Ex:
    #       REF     ALT
    #       GC      G
    #     Alt does not have to be 1 bp - any length shorter than Ref.
    # 'Insertion'
    #     Ex:
    #       REF     ALT
    #       A       ATT
    #     First base _does not_ always match REF, so there may be a SNP as well
    # 'No indel'
    #     Ex:
    #       REF     ALT
    #       A       G

    #define here, so that all sub-functions can access them
    sequences = defaultdict(dict)
    insertions = defaultdict(dict) #Currently not used, but kept in case of future use.
    metadata = {
        'meta_lines': [], # all the VCF meta_lines, i.e. those starting with '##' (they are left unparsed)
        'chrom': None,    # chromosome name (we only allow one)
        'ploidy': None,   # ploidy count -- encoded in how the GT calls are formatted
    }

    #TreeTime handles 2-3 base ambig codes, this will allow that.
    def getAmbigCode(bp1, bp2, bp3=""):
        bps = [bp1,bp2,bp3]
        bps.sort()
        key = "".join(bps)

        return {
            'CT': 'Y',
            'AG': 'R',
            'AT': 'W',
            'CG': 'S',
            'GT': 'K',
            'AC': 'M',
            'AGT': 'D',
            'ACG': 'V',
            'ACT': 'H',
            'CGT': 'B'
        }[key]

    #Parses a 'normal' (not hetero or no-call) call depending if insertion+deletion, insertion,
    #deletion, or single bp subsitution
    def parse_homozygous_call(snps, ins, pos, ref, alt):

        # Replace missing allele with N(s). See commentary in test_vcf.py::TestNoCallsOrMissing for more details
        if alt=='*' or alt=='.':
            alt = "N" * len(ref)

        #Insertion where there are also deletions (special handling)
        if len(ref) > 1 and len(alt)>len(ref):
            ## NOTE: the loop below contains a potential double-counting bug. For example,
            ## REF='TCG' ALT='TCAG' (Example 5.1.3 in the VCF 4.2 spec), then when i=2
            ## we'll add both snps[pos+2] = 'A' as well as ins[pos+2] = 'AG'. This has been 
            ## detailed within `test_vcf.py`, as it may also be the expected way to encode
            ## insertions within TreeTime
            for i in range(len(ref)):
                #if the pos doesn't match, store in sequences
                if ref[i] != alt[i]:
                    snps[pos+i] = alt[i]
                #if about to run out of ref, store rest:
                if (i+1) >= len(ref):
                    ins[pos+i] = alt[i:]
        #Deletion
        elif len(ref) > 1:
            for i in range(len(ref)):
                #if ref is longer than alt, these are deletion positions
                if i+1 > len(alt):
                    snps[pos+i] = '-'
                #if not, there may be mutations
                else:
                    if ref[i] != alt[i]:
                        snps[pos+i] = alt[i]
        #Insertion + single-base ref
        elif len(alt) > 1:
            ins[pos] = alt
            # If the first base of the allele doesn't match the ref then we _also_ have a mutation
            if ref[0]!=alt[0]:
                snps[pos] = alt[0]
        #No indel
        else:
            snps[pos] = alt


    #Parses a 'bad' (hetero or no-call) call depending on what it is
    #TODO - consider the situation where the alternate allele base(s) is '*' (done for the homozygous case)
    def parse_heterozygous_call(gen, snps, ins, pos, ref, ALT):
        #Deletion
        #   REF     ALT     Seq1    Seq2    Seq3
        #   GCC     G       1/1     0/1     ./.
        # Seq1 (processed by parseCall, above) will become 'G--'
        # Seq2 will become 'GNN'
        # Seq3 will become 'GNN'
        if len(ref) > 1:
            #Deleted part becomes Ns
            if gen[0] == '0' or gen[0] == '.':
                if gen[0] == '0':   #if het, get first bp
                    alt = str(ALT[int(gen[2])-1])
                else: #if no-call, there is no alt, so just put Ns after 1st ref base
                    alt = ref[0]
                for i in range(len(ref)):
                    #if ref is longer than alt, these are deletion positions
                    if i+1 > len(alt):
                        snps[pos+i] = 'N'
                    #if not, there may be mutations
                    else:
                        if ref[i] != alt[i]:
                            snps[pos+i] = (alt[i] if alt[i] != '.' else 'N') #'.' = no-call

        #If not deletion, need to know call type
        #if het, see if proposed alt is 1bp mutation
        elif gen[0] == '0':
            alt = str(ALT[int(gen[2])-1])
            if len(alt)==1:
                #alt = getAmbigCode(ref,alt) #if want to allow ambig
                alt = 'N' #if you want to disregard ambig
                snps[pos] = alt
            #else a het-call insertion, so ignore.

        #else it's a no-call; see if all alts have a length of 1
        #(meaning a simple 1bp mutation)
        elif len(ALT)==len("".join(ALT)):
            alt = 'N'
            snps[pos] = alt
        #else a no-call insertion, so ignore.

    def validate_alt(alt):
        """
        from the VCF 4.3 spec:
        > the ALT field must be a symbolic allele, or a breakend replacement string,
        > or match the regular expression [see ALT_CHARS regex, above]
        Since we only consider GT variation, the symbolic alleles and breakends aren't
        relevant for us.

        The spec also states:
        > Tools processing VCF files are not required to preserve case in the allele String
        
        Return the uppercase allele bases (string) _or_ None if it fails validation
        """
        if ALT_CHARS.match(alt):
            return alt.upper()
        else:
            print(f"WARNING: Encountered invalid allele base(s) {alt!r}. Skipping...")
            return None

    #House code is *much* faster than pyvcf because we don't care about all info
    #about coverage, quality, counts, etc, which pyvcf goes to effort to parse
    #(and it's not easy as there's no standard ordering). Custom code can completely
    #ignore all of this.
    from Bio import SeqIO

    #Use different openers depending on whether compressed
    opn = gzip.open if vcf_file.endswith(('.gz', '.GZ')) else open

    with opn(vcf_file, mode='rt') as f:
        current_block = "meta-information" # The start of VCF files is the meta-information (lines starting with ##)
        header,samps,nsamp=None,None,None

        for line in f:
            if line.startswith("##"):
                if current_block!='meta-information':
                    raise TreeTimeError(f"Malformed VCF file {vcf_file!r} - all the meta-information (lines starting with ##) must appear at the top of the file.")
                metadata['meta_lines'].append(line.strip())
            elif line[0]=='#':
                mandatory_fields = ['CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO', 'FORMAT']
                # Note that FORMAT isn't mandatory unless genotype data is present. But every VCF we deal with has genotype data - that's the point!
                header_start = "#"+"\t".join(mandatory_fields)
                if not line.startswith(header_start):
                    raise TreeTimeError(f"Malformed VCF file {vcf_file!r} - the header line must start with {header_start}")
                if current_block!='meta-information':
                    raise TreeTimeError(f"Malformed VCF file {vcf_file!r} - the header must immediately follow the meta-information lines")
                current_block = 'header'
                header = line.strip().split('\t')
                samps = [ x.strip() for x in header[9:] ] #ensure no leading/trailing spaces
                nsamp = len(samps)
                for sample_name in samps:
                    sequences[sample_name] = {}
            else:
                if current_block!='header' and current_block!='data-lines':
                    raise TreeTimeError(f"Malformed VCF file {vcf_file!r} - the data lines must follow the header line")
                current_block='data-lines'
                l = line.strip()
                if not l: # empty line
                    continue
                dat = l.split('\t')
                chrom = str(dat[0])
                if metadata['chrom'] is None:
                    metadata['chrom'] = chrom
                elif metadata['chrom']!=chrom:
                    raise TreeTimeError(f"The VCF file {vcf_file!r} contains multiple chromosomes. TreeTime can not yet handle this.")
                pos = int(dat[1])-1 # Convert VCF 1-based to python 0-based
                REF = dat[3]
                ALT = dat[4].split(',') # List of alternate alleles (strings)
                calls = np.array(dat[9:])
                if len(calls)!=nsamp:
                    raise TreeTimeError(f"Malformed VCF file {vcf_file!r} - the data lines have different number of calls than the number of samples defined in the header")

                # Treetime only parses GT FORMAT calls. Moreover, "the first sub-field must always be the genotype (GT) if it is present"
                # according to the VCF 4.2 spec. Note that if the VCF is for one sample only then the format's a bit different, but this'll
                # raise an error above because the header differs from what we assert. Other variation (CN, BND etc) is not parsed by
                # TreeTime, only GT.
                FORMAT = dat[8]
                if not FORMAT.startswith('GT'):
                    continue

                #get samples that differ from Ref at this site
                for sname, sa in zip(samps, calls):
                    gt = sa.split(':')[0] # May be multiple colon-separated subfields, depending on the line's FORMAT

                    # `gt` is the "index" of the alternate allele for this sample.
                    # The format of `gt` is quite varied:
                    # For haploid genomes, it's simply a single int _or_ "." (meaning a call cannot be made at the locus)
                    # For polyploid genomes, the format is a list of {int, '.'} separated by "/" or "|"
                    # ('/' means unphased, '|' means phased). 
                    # If the index is an int, it's the 1-based (!) lookup index for ALT

                    # Split on the valid separators - if haploid, then we'll get a list len=1
                    gts = gt.split('|') if '|' in gt else gt.split('/')
                    ploidy = len(gts)
                    if metadata['ploidy'] is None:
                        metadata['ploidy'] = ploidy
                    elif metadata['ploidy']!=ploidy:
                        raise TreeTimeError(f"The VCF file {vcf_file!r} had genotype calls of ploidy {metadata['ploidy']} but sample {sname!r} has a genotype of ploidy {ploidy}")

                    # if polyploid, but homozygous, then treat as if haploid
                    if ploidy>1 and len(set(gts))==1:
                        gt = gts[0]

                    if gt.isdigit(): # haploid, and a call has been made (i.e. it's not gt='.')
                        gt = int(gt)
                        if gt==0:
                            continue # reference allele!
                        alt = validate_alt(ALT[gt-1]) # gt is the 1-based lookup, but ALT is 0-indexed
                        if alt:
                            parse_homozygous_call(sequences[sname],insertions[sname], pos, REF, alt)
                        continue

                    if gt=='.': # haploid "call cannot be made" identifier - replace REF with N(s)
                        for i in range(len(REF)):
                            sequences[sname][pos+i] = 'N'
                        continue

                    # ---- heterozygous polyploid call  ----
                    parse_heterozygous_call(gt, sequences[sname],insertions[sname], pos, REF, ALT)

    #Gather all variable positions
    #NOTE: this does not consider positions of insertions
    positions = set()
    for seq, muts in sequences.items():
        positions.update(muts.keys())

    num_insertions = sum([len(list(ins.keys())) for ins in insertions.values()])
    if len(positions)==0 and num_insertions==0:
        raise TreeTimeError(f"VCF file {vcf_file!r} has no data-lines which we could extract genotype information from!")

    #One or more seqs are same as ref! (No non-ref calls) So haven't been 'seen' yet
    if nsamp > len(sequences):
        missings = set(samps).difference(sequences.keys())
        for s in missings:
            sequences[s] = {}

    if ref_file:
        refSeq = SeqIO.read(ref_file, format='fasta')
        refSeq = refSeq.upper() #convert to uppercase to avoid unknown chars later
        refSeqStr = str(refSeq.seq)
    else:
        refSeqStr = None

    compress_seq = {'reference':refSeqStr,
                    'sequences': sequences,
                    'insertions': insertions,
                    'positions': sorted(positions),
                    'metadata': metadata}

    return compress_seq


def write_vcf(tree_dict, file_name, mask=None):#, compress=False):
    """
    Writes out a VCF-style file (which seems to be minimally handleable
    by vcftools and pyvcf) of the alignment. This is created from a dict
    in a similar format to what's created by :py:meth:`treetime.vcf_utils.read_vcf`

    Positions of variable sites are transformed to start at 1 to match
    VCF convention.

    Parameters
    ----------
     tree_dict: nested dict
        A nested dict with required keys of:
        'sequences': maps sampleName -> pos (0-based) -> new base (SNP)
        'reference': string of reference nuc sequence
        'positions': sorted list of 0-based positions with variation in sequences
        And optional keys:
        'inferred_const_sites': list or set, 0-based positions to skip output for.
        This input is often created by :py:meth:`treetime.TreeAnc.get_tree_dict`
        'metadata': dict of information to influence VCF formatting. Only
        the following keys are used:
        'metadata.ploidy': int. Influences how genotype calls are formatted. (default
        of 2 (diploid) used if not provided)
        'metadata.chrom': str. The chromosome name (default of '1' used if not provided)

     file_name: str
        File to which the new VCF should be written out. File names ending with
        '.gz' will result in the VCF automatically being gzipped.

    mask : optional, str of 0 or 1
        Calls at these positions will be skipped
    """

#   Programming Logic Note:
#
#    For a sequence like:
#    Pos     1 2 3 4 5 6
#    Ref     A C T T A C
#    Seq1    A C - - - G
#
#    In a dict it is stored:
#    Seq1:{3:'-', 4:'-', 5:'-', 6:'G'}  (Numbering from 1 for simplicity)
#
#    In a VCF it needs to be:
#    POS REF     ALT     Seq1
#    2   CTTA    C       1/1
#    6   C       G       1/1
#
#    If a position is deleted (pos 3), need to get invariable position preceeding it
#
#    However, in alternative case, the base before a deletion is mutant, so need to check
#        that next position isn't a deletion (as otherwise won't be found until after the
#        current single bp mutation is written out)
#
#    When deleted position found, need to gather up all adjacent mutant positions with deletions,
#        but not include adjacent mutant positions that aren't deletions (pos 6)
#
#    Don't run off the 'end' of the position list if deletion is the last thing to be included
#        in the VCF file

    sequences = tree_dict['sequences']
    ref = tree_dict['reference']
    positions = tree_dict['positions']
    ploidy = tree_dict.get('metadata', {}).get('ploidy', 2)
    chrom_name = tree_dict.get('metadata', {}).get('chrom', '1')
    sample_names = list(sequences.keys())
    inferred_const_sites = set(tree_dict.get('inferred_const_sites', []))

    # For every variable site in sequences, flip the format around so
    # we can have fast lookups later on.
    alleles = {}
    num_samples = len(sample_names)
    for idx, name in enumerate(sample_names):
        for posn, allele in sequences[name].items():
            if posn not in alleles:
                alleles[posn] = np.zeros(num_samples, dtype='U')
            alleles[posn][idx] = allele
    # fill in reference
    for posn,bases in alleles.items():
        bases[bases==''] = ref[posn]

    def handleDeletions(i, pi, pos, ref, delete, pattern):
        refb = ref[pi]
        if delete: #Need to get the position before
            i-=1    #As we'll next go to this position again
            pi-=1
            pos = pi+1
            refb = ref[pi]
            #re-get pattern
            pattern = []
            for k,v in sequences.items():
                try:
                    pattern.append(sequences[k][pi])
                except KeyError:
                    pattern.append(ref[pi])
            pattern = np.array(pattern).astype('U')

        sites = []
        sites.append(pattern)

        #Gather all positions affected by deletion - but don't run off end of position list
        while (i+1) < len(positions) and positions[i+1] == pi+1:
            i+=1
            pi = positions[i]
            pattern = []
            for k,v in sequences.items():
                try:
                    pattern.append(sequences[k][pi])
                except KeyError:
                    pattern.append(ref[pi])
            pattern = np.array(pattern).astype('U')

            #Stops 'greedy' behaviour from adding mutations adjacent to deletions
            if any(pattern == '-'): #if part of deletion, append
                sites.append(pattern)
                refb = refb+ref[pi]
            else: #this is another mutation next to the deletion!
                i-=1    #don't append, break this loop

        #Rotate them into 'calls'
        align = np.asarray(sites).T

        #Get rid of '-', and put '0' for calls that match ref
        #Only removes trailing '-'. This breaks VCF convension, but the standard
        #VCF way of handling this* is really complicated, and the situation is rare.
        #(*deletions and mutations at the same locations)
        fullpat = []
        for pt in align:
            gp = len(pt)-1
            while pt[gp] == '-':
                pt[gp] = ''
                gp-=1
            pat = "".join(pt)
            if pat == refb:
                fullpat.append('0')
            else:
                fullpat.append(pat)

        pattern = np.array(fullpat)

        return i, pi, pos, refb, pattern


    #prepare the header of the VCF & write out
    header=["#CHROM","POS","ID","REF","ALT","QUAL","FILTER","INFO","FORMAT"]+sample_names

    opn = gzip.open if file_name.endswith(('.gz', '.GZ')) else open
    out_file = opn(file_name, 'w')

    out_file.write( "##fileformat=VCFv4.2\n"+
                        "##source=NextStrain\n"+
                        "##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">\n"+
                        f"##contig=<ID={chrom_name}>\n")
    out_file.write("\t".join(header)+"\n")

    vcfWrite = []
    errorPositions = []
    explainedErrors = 0
    mask_skip_count = 0

    #Why so basic? Because we sometimes have to back up a position!
    i=0
    while i < len(positions):
        #Get the 'pattern' of all calls at this position.
        #Look out specifically for current (this pos) or upcoming (next pos) deletions
        #But also distinguish these two, as handled differently.

        pi = positions[i]
        pos = pi+1 #change numbering to match VCF, not python, for output
        refb = ref[pi] #reference base at this position
        delete = False #deletion at this position - need to grab previous base (invariable)
        deleteGroup = False #deletion at next position (mutation at this pos) - do not need to get prev base

        if mask and mask[pi] == '1':
            mask_skip_count+=1
            i+=1
            continue

        # patterns will be empty if there was no variation in the sequences
        pattern = alleles[pi] if pi in alleles else np.array([]).astype('U')
        pattern2 = alleles[pi+1] if pi+1 in alleles else np.array([]).astype('U')

        #If a deletion here, need to gather up all bases, and position before
        if any(pattern == '-'):
            if pos != 1:
                deleteGroup = True
                delete = True
            else:
                #If theres a deletion in 1st pos, VCF files do not handle this well.
                #Proceed keeping it as '-' for alt (violates VCF), but warn user to check output.
                #(This is rare)
                print(fill("WARNING: You have a deletion in the first position of your"
                           " alignment. VCF format does not handle this well. Please check"
                           " the output to ensure it is correct."))
        else:
            #If a deletion in next pos, need to gather up all bases
            if any(pattern2 == '-'):
                deleteGroup = True

        #If deletion, treat affected bases as 1 'call':
        if delete or deleteGroup:
            if pattern.size==0: # no variation in sequences
                pattern = np.full(num_samples, refb, dtype='U')
            i, pi, pos, refb, pattern = handleDeletions(i, pi, pos, ref, delete, pattern)
        #If no deletion, replace ref with '0' which means the reference base is unchanged
        else:
            pattern[pattern==refb] = '0'

        #Get the list of ALTs - minus any '0' which are unchanged reference sequences!
        uniques = np.unique(pattern)
        uniques = uniques[np.where(uniques!='0')]

        #Convert bases to the number that matches the ALT
        j=1
        for u in uniques:
            pattern[np.where(pattern==u)[0]] = str(j)
            j+=1

        #What if there's no variation at a variable site??
        #This can happen when sites are modified by TreeTime - see below.
        #We don't print to VCF (because no variation!)
        any_variation = len(uniques)!=0
        if not any_variation:
            #If we expect it (it was made constant by TreeTime), it's fine.
            if pi in inferred_const_sites:
                explainedErrors += 1
            else:
                #If we don't expect, raise an error
                errorPositions.append(str(pi))

        #Write it out - Increment positions by 1 so it's in VCF numbering
        #If no longer variable, and explained, don't write it out
        if any_variation:
            calls = [ "/".join([j]*ploidy) for j in pattern ]
            output = [chrom_name, str(pos), ".", refb, ",".join(uniques), ".", "PASS", ".", "GT"] + calls
            vcfWrite.append("\t".join(output))

        i+=1

    #Note: The number of 'inferred_const_sites' passed back by TreeTime will often be longer
    #than the number of 'site that were made constant' that prints below. This is because given the site:
    # Ref   Alt     Seq
    # G     A       AANAA
    #This will be converted to 'AAAAA' and listed as an 'inferred_const_sites'. However, for VCF
    #purposes, because the site is 'variant' against the ref, it is variant, as expected, and so
    #won't be counted in the below list, which is only sites removed from the VCF.

    if mask_skip_count:
        print(f"{mask_skip_count} positions were skipped due to the provided mask")

    if 'inferred_const_sites' in tree_dict and explainedErrors != 0:
        print(fill("Sites that were constant except for ambiguous bases were made" +
                   " constant by TreeTime. This happened {} times. These sites are".format(explainedErrors) +
                   " now excluded from the VCF."))

    if len(errorPositions) != 0:
        print ("\n***WARNING: vcf_utils.py")
        print(fill("\n{} sites were found that had no alternative bases.".format(str(len(errorPositions)))+
                  " If this data has been run through TreeTime and contains ambiguous bases,"
                  " try calling get_tree_dict with var_ambigs=True to see if this clears the error."))
        print(fill("\nAlternative causes:"
                   "\n- Not all sequences in your alignment are in the tree"
                  " (if you are running TreeTime via commandline this is most likely)"
                  "\n- In TreeTime, can be caused by overwriting variants in tips with small branch lengths (debug)"
                  "\n\nThese are the positions affected (numbering starts at 0):"))
        print(fill(", ".join(errorPositions)))

    out_file.write("\n".join(vcfWrite))
    out_file.close()


def process_sparse_alignment(aln, ref, ambiguous_char):
    return process_alignment_dictionary(aln, ref, ambiguous_char)

def process_alignment_dictionary(aln, ref, ambiguous_char):
    """
    prepare the dictionary specifying differences from a reference sequence
    to construct the reduced alignment with variable sites only. NOTE:
        - sites can be constant but different from the reference
        - sites can be constant plus a ambiguous sites

    assigns
    -------
    - self.nonref_positions: at least one sequence is different from ref

    Returns
    -------
    reduced_alignment_const
        reduced alignment accounting for non-variable postitions

    alignment_patterns_const
        dict pattern -> (pos in reduced alignment, list of pos in full alignment)

    variable_positions
        list of variable positions needed to construct remaining

    """

    # number of sequences in alignment
    nseq = len(aln)

    inv_map = defaultdict(list)
    for k,v in aln.items():
        for pos, bs in v.items():
            inv_map[pos].append(bs)

    nonref_positions = np.sort(list(inv_map.keys()))
    constant_up_to_ambiguous = []

    nonref_const = []
    nonref_alleles = []
    ambiguous_const = []
    variable_pos = []
    for pos, bs in inv_map.items(): #loop over positions and patterns
        bases = list(np.unique(bs))
        if len(bs) == nseq: #every sequence is different from reference
            if (len(bases)<=2 and ambiguous_char in bases) or len(bases)==1:
                # all sequences different from reference, but only one state
                # (other than ambiguous_char) in column
                nonref_const.append(pos)
                if len(bases)==1:
                    nonref_alleles.append(bases[0])
                else:
                    nonref_alleles.append([x for x in bases if x!=ambiguous_char][0])

                if ambiguous_char in bases: #keep track of sites 'made constant'
                    constant_up_to_ambiguous.append(pos)
            else:
                # at least two non-reference alleles
                variable_pos.append(pos)
        else: # not every sequence different from reference
            if len(bases)==1 and bases[0]==ambiguous_char:
                ambiguous_const.append(pos)
                constant_up_to_ambiguous.append(pos) #keep track of sites 'made constant'
            else:
                # at least one non ambiguous non-reference allele not in
                # every sequence
                variable_pos.append(pos)

    refMod = np.copy(ref)
    # place constant non reference positions by their respective allele
    refMod[nonref_const] = nonref_alleles
    # mask variable positions
    states = np.unique(refMod)
    refMod[variable_pos] = '.'

    # for each base in the gtr, make constant alignment pattern and
    # assign it to all const positions in the modified reference sequence
    constant_columns = []
    constant_patterns = {}
    for base in states:
        if base==ambiguous_char:
            continue
        p = np.repeat(base, nseq)
        pos = list(np.where(refMod==base)[0])
        #if the alignment doesn't have a const site of this base, don't add! (ex: no '----' site!)
        if len(pos):
            constant_patterns["".join(p.astype('U'))] = [len(constant_columns), pos]
            constant_columns.append(p)

    return {"constant_columns": constant_columns,
            "constant_patterns": constant_patterns,
            "variable_positions": variable_pos,
            "nonref_positions": nonref_positions,
            "constant_up_to_ambiguous": constant_up_to_ambiguous}


